#!/usr/bin/perl -I /usr/local/lib/perl/site_perl
# A perl-based backup script: Read docco 
$|=1;

=head1 NAME
 threaded-backup.pl
=cut

=head1 DECRIPTION
 A perl-based backup script. Configuration and logging data are held in an sqlite3 database.
 A query is run against this database to determine which machines to back up. Machines to
 back up are sent through queues to writer-threads, one for each VG, where the heavy lifting
 happens.
=cut

use strict;
use threads;
use LockFile::Simple;
use Thread::Queue;
use DBI;
use Net::Ping;
use JSON;

use ZFS::Host;
use ZFS::Config;

# threaded-backup.h
sub getvgs();    # Get a list of VGs
sub topsighup(); # Handle a HUP at the top level
sub topsighup(); # Handle an INT at the top level
sub tidydatabase(); # Remove defunct entries from the database due to crashes/TERMs

# Settings:
my $dbfile:shared='/usr/local/backupsv2/etc/config.db3'; # Where is the database file stored?

# Constants: M_ - messages; D_ - destinations; SLEEP
use constant {
 D_MAIN=>0x01,
 D_CRON=>0x02,
 D_DISPATCHER=>0x04,
 D_WRITER=>0x08,
 D_MASK=>0x0f,

 M_STOP=>0x10,
 M_MASK=>0xf0,

 SLEEP_WATCHER=>1,
 SLEEP_WRITER=>1,
 SLEEP_CRON=>2,
};

# Check we're not running in parallel
my $lockmgr=LockFile::Simple->make(-autoclean=>1, -max=>1, -hold=>0, -ext=>'.pid');
$lockmgr->lock("/var/run/threaded-backup") || die "Cannot lock /var/run/threaded-backup.pid\n";
# Remove any interrupted jobs. 
&tidydatabase;

# A queue of messages broadcast to all threads.
my $controlq: shared=Thread::Queue->new();

# Create some writers
my $writers={};
my $writerqs={};

# Handle singals: 
#  HUP does a graceful shutdown; 
#  INT doesn't yet do an immediate shutdown
$SIG{HUP}=\&topsighup;
$SIG{INT}=\&topsigint;

# We shall create one writer thread for each zpool
my @zpools=&getzpools();
foreach my $zpool (@zpools) {
 my $tq:shared=Thread::Queue->new();
 $writers->{$zpool}=threads->create('writer',$zpool,$tq);
 $writerqs->{$zpool}=$tq;
}

# And one thread to control the writers
my $cron=threads->create('cron',$writerqs);

# As long as we have all those threads, just keep on running
while (scalar(threads->list(threads::running))==scalar(@zpools)+1) {
  sleep(2);
}

# Close-down code
print "Shutting down: only ".scalar(threads->list(threads::running))." threads running\n";
# Send cron the stop signal if it's still running
$controlq->enqueue(D_CRON|M_STOP) if $cron->is_running();
# Send writers the stop signal
$controlq->enqueue(D_WRITER|M_STOP);

# Wait for all threadsA
$SIG{HUP}='DEFAULT';
$SIG{INT}='DEFAULT';
$cron->join();
foreach my $zpool (@zpools) { $writers->{$zpool}->join();}
exit;

# Handle SIGHUP: send stop messages to all threads and stop cleanly.
sub topsighup() {
 print "Gracefully terminating\n";
 $controlq->enqueue(D_CRON|M_STOP) if $cron->is_running();
 $controlq->enqueue(D_WRITER|M_STOP);
 print "Stop messages sent\n";
}

# Handle SIGINT: send INT signals to all threads and stop immediately.
sub topsigint() {
 print "INT caught: stopping\n";
 $cron->kill('SIGINT') if defined($cron);
 foreach my $zpool (keys(%{$writers})) { $writers->{$zpool}->kill('SIGINT');};
}

# Thread which determines which machines to backup.
sub cron {
 my $qs=shift;
 print "Cron thread started\n";
 # Handle INTerrupt signal
 $SIG{'INT'}=sub { print "Cron INTd\n";threads->exit();$SIG{'INT'}='DEFAULT';};
 my $keepgoing=1;
 my $ping=Net::Ping->new();
 my $dbh=DBI->connect("dbi:SQLite:dbname=$dbfile","","");
 my $sql='select backup_task_id,hostname,zfs_target from backup_queue';
 my $offlinesql='update host set offline=1 where hostname=?';
 my $sth=$dbh->prepare($sql);
 my $off_sth=$dbh->prepare($offlinesql);
 while ($keepgoing) {
  $sth->execute() || die "Unable to run $sql: ".$dbh->errstr;
  # Get any items we need to backup
  while (my $row=$sth->fetchrow_hashref) {
   my $h=$row->{'hostname'};
   if ($ping->ping($h)) {
    my $zpool=$row->{'zfs_target'};
    $zpool =~ s/\/.*//;
    # Assign to a queue
    my $q=$qs->{$zpool};
    if (defined($q)) {
     my $logid=&stampjoinqueue($dbh,$row->{'backup_task_id'});
     #print "Enqueueing ".&shorthost($h)." item id ".$row->{'backup_item_id'}." with log id ".$logid."\n";
     # Queue the item by adding the unique log ID
     $q->enqueue($logid);
    } else {
     warn "Unable to enqueue entry for $h: Could not determine a VG";
    }
   } else {
    # Machine is offline
    print "Machine $h is offline\n";
    $off_sth->execute($h);
   }
  }
  # Should we quit?
  if (examinecontrolq(D_CRON) & M_STOP) {
   $keepgoing=0;
  } else {
   sleep(SLEEP_CRON);
   threads->exit() unless $dbh->ping() 
  }
 }
 print "Cron thread ending\n";
}

# Utility function to get the short name from an FQDN
sub shorthost($) {
 my $hn=shift;
 $hn =~ s/\..*//;
 return $hn;
}

# Utility function to decide whether a host is alive
sub ping($) {
 my $hn=shift;
 my $p=Net::Ping->new();
 return $p->ping($hn);
}

# A thread to receive jobs from the queue and run them
sub writer($$) {
 my $vg=shift;
 my $q=shift;
 $SIG{'INT'}=sub { print "writer $vg INTd\n";threads->exit();$SIG{'INT'}='DEFAULT';};
 my $dbh=DBI->connect("dbi:SQLite:dbname=$dbfile","","");
 my $keepgoing=1;
 my $actions={zfs => \&zfswrite, zfs_rsync=> \&zfsrsyncwrite};
 while ($keepgoing) {
  # Get a structure from the queue
  my $qlen=$q->pending();
  if ($qlen) {
   my @items=$q->dequeue_nb();
   my $sql="select backup_type_name from writer_view where backup_log_id=?";
   my $sth=$dbh->prepare($sql);
   foreach my $id (@items) {
    $sth->execute($id);
    # Really ought to only get one row at a time
    while (my $row=$sth->fetchrow_hashref) {
     #backup_log_id backup_item_id backup_class hostname timeout arguments age
     my $class=$row->{'backup_type_name'};
     $actions->{$class}->($dbh,$id);
    }
   }
  }
  # Should we quit?
  if (examinecontrolq(D_WRITER) & M_STOP) {
   $keepgoing=0;
  } else {
   sleep(SLEEP_WRITER);
  }
 }
 print "Writer for $vg terminating\n";
}

sub dummyexecute($$$) {
 my $dbh=shift;
 my $hostname=shift;
 my $id=shift;
 my $aref=shift;
 #use Data::Dumper;
 #print "Dummy writer : will write a backup for $hostname".Dumper($aref)."\n";
}

sub zfswrite($$) {
 my ($dbh,$id)=@_;
 print "ZFS writer requested for job $id\n";
 # Look up some details about this job
 my $sql='select * from zfs_detail_view where backup_log_id=?';
 my $args=$dbh->selectrow_hashref($sql,undef,$id);
 print "ZFS writer received job $id for ".$args->{'hostname'}."\n";
 &stampbackupstart($dbh,$id);
 print "zfs sending ".$args->{'zfs_source'}." to ".$args->{'zfs_target'}."\n";
 # TODO: Write log file somewhere sensible like /var/log/backup/$hostname-$i
 my $fh;
 print "/usr/local/backupsv2/bin/zfs-backup.sh $args->{'hostname'} $args->{'zfs_source'} $args->{'zfs_target'} $args->{'recurse'}\n";
 system ('/usr/local/backupsv2/bin/zfs-backup.sh',$args->{'hostname'},$args->{'zfs_source'},$args->{'zfs_target'},$args->{'recurse'});

 my $ret=$?;
 &stampbackupstop($dbh,$id,$ret);
 print "ZFS writer completed job $id for ".$args->{'hostname'}." - ret code $ret\n";
 &zfstidyup($dbh,$id,$args->{'zfs_target'});
}

sub zfsrsyncwrite($$) {
 my ($dbh,$id)=@_;
 print "ZFS rsync writer requested for job $id\n";
 # Look up some details about this job
 my $sql='select * from zfs_rsync_detail_view where backup_log_id=?';
 my $args=$dbh->selectrow_hashref($sql,undef,$id);
 print "ZFS rsync writer received job $id for ".$args->{'hostname'}."\n";
 &stampbackupstart($dbh,$id);
 print "zfs rsyncing ".$args->{'directory_source'}." to ".$args->{'zfs_target'}."\n";
 # TODO: Write log file somewhere sensible
 my $hf;
 print "/usr/local/backupsv2/bin/zfs-rsync.sh $args->{'hostname'} $args->{'directory_source'} $args->{'zfs_target'}\n";
 system('/usr/local/backupsv2/bin/zfs-rsync.sh',$args->{'hostname'},$args->{'directory_source'},$args->{'zfs_target'});
 my $ret=$?;
 &stampbackupstop($dbh,$id,$ret);
 print "ZFS rsync writer completed job $id for ".$args->{'hostname'}." returning $ret\n";
 &zfstidyup($dbh,$id,$args->{'zfs_target'});
}

sub zfstidyup($$$) {
 my $dbh=shift;
 my $id=shift;
 my $zfs=shift;
 my $zfshost=new ZFS::Host('localhost');
 print "Tidying up snapshots on $zfs\n";
 my $sql='select * from zfs_pruning_view where backup_log_id=?';
 my $arr=$dbh->selectall_hashref($sql,'flagname',undef,$id);
 my %archivepolicy=%$arr;
 my $config=ZFS::Config->void();
 $config->archivepolicy($arr);
 
 $zfshost->zfsName($zfs);
 my $list=$zfshost->getListForPrune();
 my $mostrecent=$zfshost->getLastSnapTime();
 #print "Config\n";
 #print Dumper($config);
 foreach my $type ($config->getArchiveTypes) {
  #print "Preparing $type pass\n";
  #print "Config is:\n";
  #print Dumper($config);
  $list=&addtag($type,\$mostrecent,\$config,\$list,\$zfshost);
  #print "Finished $type pass\n";
  #print Dumper($list);
 }
#print "List\n";
#print Dumper($list);
#print "ZFShost\n";
#print Dumper($zfshost);
#print "Calling delete untagged\n";
 &deleteuntagged(\$list,\$zfshost);


 #print Dumper($list);
 #use Data::Dumper;
 #print Dumper($zfshost);
 print "Tidied\n";
}

# Indicate when a host has joined a queue. The returned ID identifies this backup job
sub stampjoinqueue($$) {
 my ($dbh,$id)=@_;
 my $sql="insert into backup_log (backup_task_id, joined_queue) values (?,datetime('now','localtime'))";
 $dbh->do($sql,undef,$id);
 $sql="select backup_log_id from backup_log where backup_task_id=? order by joined_queue desc limit 1";
 my @a=$dbh->selectrow_array($sql,undef,$id);
 return $a[0];
}

# Indicate when a backup started
sub stampbackupstart($$) {
 my ($dbh,$id)=@_;
 my $sql="update backup_log set started_processing=datetime('now','localtime') where backup_log_id=?";
 $dbh->do($sql,undef,$id);
}

# Indicate when a backup has ended
sub stampbackupstop($$$) {
 my ($dbh,$id,$exit)=@_;
 my $sql="update backup_log set ended_processing=datetime('now','localtime'), exit_code=? where backup_log_id=?";
 $dbh->do($sql,undef,$exit,$id);
}

sub rsnapshotextrainfo($$$) {
 my ($dbh,$id,$type)=@_;
 print "Updating backup_log_rsnapshot backup_log_id=$id, type=$type\n";
 my $sql="insert into backup_log_rsnapshot (backup_log_id,rsnapshot_subtype_id) values (?,(select id from rsnapshot_subtypes where desc=?))"; 
 $dbh->do($sql,undef,$id,$type);
}

# Remove incomplete log entries
sub tidydatabase() {
 my $dbh=DBI->connect("dbi:SQLite:dbname=$dbfile","","");
 #my $sql='delete from backup_log where offline !=1 or offline is null and ended_processing is null';
 my $sql='delete from backup_log where ended_processing is null and (offline =0 or offline is null)';
 $dbh->do($sql,undef);
}

sub setoffline($$) {
 my ($dbh,$id)=@_;
 #my $dbh=DBI->connect("dbi:SQLite:dbname=$dbfile","","");
 my $sql='update backup_log set offline=1 where backup_log_id=?';
 $dbh->do($sql,undef,$id);
}

# Examine control queue
sub examinecontrolq($) {
 my $dmask=shift;
 my $return=0;
 {
  lock($controlq);
  if ($controlq->pending()) {
   my $msg=$controlq->peek();
   if (($msg & D_MASK) == $dmask) {
    $msg=$controlq->dequeue();
   }
   $return=($msg & M_MASK);
  }
 }
 return $return;
}

sub getzpools() {
 my @zpools=`zpool list -H -o name`;
 chomp(@zpools);
 return @zpools;
}

sub addtag($$$$$) {
        #my ($type,$mostrecent,$config,$list,$host)=@_;
        my $type=shift;
	my $mostrecent=shift;
	my $rconfig=shift;
	my $rlist=shift;
	my $rhost=shift;
	my $config=$$rconfig;
	my $list=$$rlist;
	my $host=$$rhost;
    
        #print "Investigating $type tags\n";
        #print Dumper($list);
        #print "Most recent: $mostrecent\n";
#print "Config in AddTag\n";
#print Dumper($config);
        my $maxage=$mostrecent-$config->ageOfArchiveType($type);
        my $step=$config->ageOfArchiveType($type)/$config->numberOfArchiveType($type);
        my $c=&counttags($type,$list);
        # Bootstrap code
        if ($c==0) {
                # Find oldest snapshot < maximum age
                #print "Bootstrapping $type: need candidates with stamp>$maxage\n";
                my @candidates=sort(grep {$_>=($maxage)} keys(%{$list}));
                my $bootstrap=$candidates[0];
                push(@{$list->{$bootstrap}},$type);
                $host->holdSnapshot($type,$bootstrap);
                #print "Bootstrapping $type by tagging $bootstrap\n";
                #print Dumper($list);
        }
        # Algorithm code: Find newest tagged snapshot
        my $leftmost=&newest($type,$list);
        #print "Located $leftmost tagged as $type \n";
        # Find any snapshots to add
        #print "Seeking others newer than ".($leftmost+$step)." $step away\n";
        my $nextsnap=&newerthan($leftmost+$step,$list);
        if ($nextsnap) { 
                #print "Found $nextsnap - tagging\n";
                push(@{$list->{$nextsnap}},$type);
                $host->holdSnapshot($type,$nextsnap);
                #print Dumper($list);
                &addtag($type,$mostrecent,\$config,\$list,\$host);
        }
        # De-tag any snapshots which are older than $maxage
        my @candidates=grep {$_<$maxage} keys(%{$list});
        foreach my $k (sort(@candidates)) {
                my $c=scalar(@{$list->{$k}});
                #print "$k has $c entries\n";
                if (scalar(@{$list->{$k}})>0) {
                        $host->releaseSnapshot($type,$k) if grep(/$type/, @{$list->{$k}});
#print "removing $type from $k : ".join(', ',@{$list->{$k}})."\n";
                        my @l=grep {!/$type/} @{$list->{$k}};
                        $list->{$k}=[];
                        push(@{$list->{$k}},@l);
#print "removed $type from $k : ".join(', ',@{$list->{$k}})."\n";
                }
        }
	return $list;
}

sub deleteuntagged($$) {
	#my ($list,$host)=shift;
	my $rlist=shift;
	my $list=$$rlist;
	my $rhost=shift;
	my $host=$$rhost;
#print "In deleteuntagged : list\n";
#print Dumper($list);
#print "--HOST---\n";
#print Dumper($host);
# Remove any untagged snapshots
        foreach my $k (keys(%{$list})) {
                if (scalar(@{$list->{$k}})==0) {
                        delete $list->{$k};
                        $host->destroySnapshot($k);
                }
        }
}


sub newerthan($$) {
        my ($age,$list)=@_;
        my @list=(sort(grep {$_>=$age} keys(%{$list})));

        return $list[0];
}

sub newest($$) {
        my ($type,$list)=@_;
        my @list=reverse(sort(grep { grep /$type/,@{$list->{$_}};} keys(%{$list})));
        return $list[0];
}

sub counttags($$) {
        my ($type,$list)=@_;
        #my $count=0;
        my $count=scalar(grep { grep /$type/,@{$list->{$_}};} keys(%{$list}));
        #print "Found $count $type items\n";
        return $count;
}

